#include "interface.h"

#include <fstream>

void interface_addFile(
        const char * ,
        const char * ,
        const char * ,
        size_t ) {
}

void interface_loadAllFiles() {
}

void interface_shareFile(
        const char * ,
        const char * ,
        const char * ,
        size_t ) {
}

void interface_openFile(
        const char * ,
        const char * ,
        const char * ,
        size_t ) {
}

void interface_deleteFile(
        const char * ,
        const char * ) {
}

void interface_receiveFile(
        const char * ,
        const char * ,
        const char * ,
        size_t ) {
}

bool interface_needReloadFiles() {
    return false;
}
