#ifndef interface_h
#define interface_h

#ifdef __cplusplus
#include "common.h"

extern "C" {
#endif

void updateMain();

float getStatusBarHeightDevice();

#ifdef __cplusplus
}
#endif

#endif /* interface_h */
